@@ -3,6 +3,7 @@
 package org.torproject.descriptor.impl;
 
 import org.torproject.descriptor.DescriptorParseException;
+
 import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.HashSet;
@@ -42,14 +43,15 @@ public class RelayNetworkStatusVoteImpl extends NetworkStatusImpl
       throws DescriptorParseException {
     super(voteBytes, failUnrecognizedDescriptorLines, false, false);
     Set<String> exactlyOnceKeywords = new HashSet<>(Arrays.asList((
-        "vote-status,consensus-methods,published,valid-after,fresh-until,"
+        "vote-status,published,valid-after,fresh-until,"
         + "valid-until,voting-delay,known-flags,dir-source,"
         + "dir-key-certificate-version,fingerprint,dir-key-published,"
         + "dir-key-expires,dir-identity-key,dir-signing-key,"
         + "dir-key-certification,directory-signature").split(",")));
     this.checkExactlyOnceKeywords(exactlyOnceKeywords);
     Set<String> atMostOnceKeywords = new HashSet<>(Arrays.asList((
-        "client-versions,server-versions,flag-thresholds,params,contact,"
+        "consensus-methods,client-versions,server-versions,"
+        + "flag-thresholds,params,contact,"
         + "legacy-key,dir-key-crosscert,dir-address,directory-footer").
         split(",")));
     this.checkAtMostOnceKeywords(atMostOnceKeywords);
@@ -74,7 +76,8 @@ public class RelayNetworkStatusVoteImpl extends NetworkStatusImpl
     this.ignoringAdvertisedBws = -1;
 
     Scanner s = new Scanner(new String(headerBytes)).useDelimiter("\n");
-    boolean skipCrypto = false; /* TODO Parse crypto parts. */
+    String nextCrypto = "";
+    StringBuilder crypto = null;
     while (s.hasNext()) {
       String line = s.next();
       String[] parts = line.split("[ \t]+");
@@ -144,18 +147,52 @@ public class RelayNetworkStatusVoteImpl extends NetworkStatusImpl
         this.parseDirKeyExpiresLine(line, parts);
         break;
       case "dir-identity-key":
+        this.parseDirIdentityKeyLine(line, parts);
+        nextCrypto = "dir-identity-key";
+        break;
       case "dir-signing-key":
+        this.parseDirSigningKeyLine(line, parts);
+        nextCrypto = "dir-signing-key";
+        break;
       case "dir-key-crosscert":
+        this.parseDirKeyCrosscertLine(line, parts);
+        nextCrypto = "dir-key-crosscert";
+        break;
       case "dir-key-certification":
+        this.parseDirKeyCertificationLine(line, parts);
+        nextCrypto = "dir-key-certification";
         break;
       case "-----BEGIN":
-        skipCrypto = true;
+        crypto = new StringBuilder();
+        crypto.append(line).append("\n");
         break;
       case "-----END":
-        skipCrypto = false;
+        crypto.append(line).append("\n");
+        String cryptoString = crypto.toString();
+        crypto = null;
+        switch (nextCrypto) {
+        case "dir-identity-key":
+          this.dirIdentityKey = cryptoString;
+          break;
+        case "dir-signing-key":
+          this.dirSigningKey = cryptoString;
+          break;
+        case "dir-key-crosscert":
+          this.dirKeyCrosscert = cryptoString;
+          break;
+        case "dir-key-certification":
+          this.dirKeyCertification = cryptoString;
+          break;
+        default:
+          throw new DescriptorParseException("Unrecognized crypto "
+              + "block in vote.");
+        }
+        nextCrypto = "";
         break;
       default:
-        if (!skipCrypto) {
+        if (crypto != null) {
+          crypto.append(line).append("\n");
+        } else {
           if (this.failUnrecognizedDescriptorLines) {
             throw new DescriptorParseException("Unrecognized line '"
                 + line + "' in vote.");
@@ -343,6 +380,7 @@ public class RelayNetworkStatusVoteImpl extends NetworkStatusImpl
       throw new DescriptorParseException("Illegal hostname in '" + line
           + "'.");
     }
+    this.hostname = parts[3];
     this.address = ParseHelper.parseIpv4Address(line, parts[4]);
     this.dirPort = ParseHelper.parsePort(line, parts[5]);
     this.orPort = ParseHelper.parsePort(line, parts[6]);
@@ -412,6 +450,34 @@ public class RelayNetworkStatusVoteImpl extends NetworkStatusImpl
         parts, 1, 2);
   }
 
+  private void parseDirIdentityKeyLine(String line, String[] parts)
+      throws DescriptorParseException {
+    if (!line.equals("dir-identity-key")) {
+      throw new DescriptorParseException("Illegal line '" + line + "'.");
+    }
+  }
+
+  private void parseDirSigningKeyLine(String line, String[] parts)
+      throws DescriptorParseException {
+    if (!line.equals("dir-signing-key")) {
+      throw new DescriptorParseException("Illegal line '" + line + "'.");
+    }
+  }
+
+  private void parseDirKeyCrosscertLine(String line, String[] parts)
+      throws DescriptorParseException {
+    if (!line.equals("dir-key-crosscert")) {
+      throw new DescriptorParseException("Illegal line '" + line + "'.");
+    }
+  }
+
+  private void parseDirKeyCertificationLine(String line, String[] parts)
+      throws DescriptorParseException {
+    if (!line.equals("dir-key-certification")) {
+      throw new DescriptorParseException("Illegal line '" + line + "'.");
+    }
+  }
+
   protected void parseFooter(byte[] footerBytes)
       throws DescriptorParseException {
     Scanner s = new Scanner(new String(footerBytes)).useDelimiter("\n");
@@ -441,6 +507,11 @@ public class RelayNetworkStatusVoteImpl extends NetworkStatusImpl
     return this.identity;
   }
 
+  private String hostname;
+  public String getHostname() {
+    return this.hostname;
+  }
+
   private String address;
   public String getAddress() {
     return this.address;
@@ -481,9 +552,33 @@ public class RelayNetworkStatusVoteImpl extends NetworkStatusImpl
     return this.dirKeyExpiresMillis;
   }
 
-  private String signingKeyDigest;
+  private String dirIdentityKey;
+  public String getDirIdentityKey() {
+    return this.dirIdentityKey;
+  }
+
+  private String dirSigningKey;
+  public String getDirSigningKey() {
+    return this.dirSigningKey;
+  }
+
+  private String dirKeyCrosscert;
+  public String getDirKeyCrosscert() {
+    return this.dirKeyCrosscert;
+  }
+
+  private String dirKeyCertification;
+  public String getDirKeyCertification() {
+    return this.dirKeyCertification;
+  }
+
   public String getSigningKeyDigest() {
-    return this.signingKeyDigest;
+    String signingKeyDigest = null;
+    if (!this.directorySignatures.isEmpty()) {
+      signingKeyDigest = this.directorySignatures.get(
+          this.directorySignatures.firstKey()).getSigningKeyDigest();
+    }
+    return signingKeyDigest;
   }
 
   private int networkStatusVersion;
@@ -493,7 +588,8 @@ public class RelayNetworkStatusVoteImpl extends NetworkStatusImpl
 
   private Integer[] consensusMethods;
   public List<Integer> getConsensusMethods() {
-    return Arrays.asList(this.consensusMethods);
+    return this.consensusMethods == null ? null :
+        Arrays.asList(this.consensusMethods);
   }
 
   private long publishedMillis;
