@@ -18,42 +18,28 @@ import org.torproject.descriptor.BandwidthHistory;
 import org.torproject.descriptor.DescriptorParseException;
 import org.torproject.descriptor.ServerDescriptor;
 
-/* Contains a relay server descriptor. */
-public class ServerDescriptorImpl extends DescriptorImpl
+/* Contains a server descriptor. */
+public abstract class ServerDescriptorImpl extends DescriptorImpl
     implements ServerDescriptor {
 
-  protected static List<ServerDescriptor> parseDescriptors(
-      byte[] descriptorsBytes, boolean failUnrecognizedDescriptorLines)
-      throws DescriptorParseException {
-    List<ServerDescriptor> parsedDescriptors =
-        new ArrayList<ServerDescriptor>();
-    List<byte[]> splitDescriptorsBytes =
-        DescriptorImpl.splitRawDescriptorBytes(descriptorsBytes,
-        "router ");
-    for (byte[] descriptorBytes : splitDescriptorsBytes) {
-      ServerDescriptor parsedDescriptor =
-          new ServerDescriptorImpl(descriptorBytes,
-          failUnrecognizedDescriptorLines);
-      parsedDescriptors.add(parsedDescriptor);
-    }
-    return parsedDescriptors;
-  }
-
   protected ServerDescriptorImpl(byte[] descriptorBytes,
       boolean failUnrecognizedDescriptorLines)
       throws DescriptorParseException {
     super(descriptorBytes, failUnrecognizedDescriptorLines, false);
     this.parseDescriptorBytes();
     this.calculateDigest();
-    Set<String> exactlyOnceKeywords = new HashSet<String>(Arrays.asList(
+    this.calculateDigestSha256();
+    Set<String> exactlyOnceKeywords = new HashSet<>(Arrays.asList(
         "router,bandwidth,published".split(",")));
     this.checkExactlyOnceKeywords(exactlyOnceKeywords);
-    Set<String> atMostOnceKeywords = new HashSet<String>(Arrays.asList((
-        "platform,fingerprint,hibernating,uptime,contact,family,"
-        + "read-history,write-history,eventdns,caches-extra-info,"
-        + "extra-info-digest,hidden-service-dir,protocols,"
-        + "allow-single-hop-exits,onion-key,signing-key,ipv6-policy,"
-        + "ntor-onion-key,router-signature").split(",")));
+    Set<String> atMostOnceKeywords = new HashSet<>(Arrays.asList((
+        "identity-ed25519,master-key-ed25519,platform,fingerprint,"
+        + "hibernating,uptime,contact,family,read-history,write-history,"
+        + "eventdns,caches-extra-info,extra-info-digest,"
+        + "hidden-service-dir,protocols,allow-single-hop-exits,onion-key,"
+        + "signing-key,ipv6-policy,ntor-onion-key,onion-key-crosscert,"
+        + "ntor-onion-key-crosscert,router-sig-ed25519,router-signature,"
+        + "router-digest-sha256,router-digest").split(",")));
     this.checkAtMostOnceKeywords(atMostOnceKeywords);
     this.checkFirstKeyword("router");
     if (this.getKeywordCount("accept") == 0 &&
@@ -68,7 +54,7 @@ public class ServerDescriptorImpl extends DescriptorImpl
   private void parseDescriptorBytes() throws DescriptorParseException {
     Scanner s = new Scanner(new String(this.rawDescriptorBytes)).
         useDelimiter("\n");
-    String nextCrypto = null;
+    String nextCrypto = "";
     List<String> cryptoLines = null;
     while (s.hasNext()) {
       String line = s.next();
@@ -79,102 +65,171 @@ public class ServerDescriptorImpl extends DescriptorImpl
           line.substring("opt ".length()) : line;
       String[] partsNoOpt = lineNoOpt.split("[ \t]+");
       String keyword = partsNoOpt[0];
-      if (keyword.equals("router")) {
+      switch (keyword) {
+      case "router":
         this.parseRouterLine(line, lineNoOpt, partsNoOpt);
-      } else if (keyword.equals("or-address")) {
+        break;
+      case "or-address":
         this.parseOrAddressLine(line, lineNoOpt, partsNoOpt);
-      } else if (keyword.equals("bandwidth")) {
+        break;
+      case "bandwidth":
         this.parseBandwidthLine(line, lineNoOpt, partsNoOpt);
-      } else if (keyword.equals("platform")) {
+        break;
+      case "platform":
         this.parsePlatformLine(line, lineNoOpt, partsNoOpt);
-      } else if (keyword.equals("published")) {
+        break;
+      case "published":
         this.parsePublishedLine(line, lineNoOpt, partsNoOpt);
-      } else if (keyword.equals("fingerprint")) {
+        break;
+      case "fingerprint":
         this.parseFingerprintLine(line, lineNoOpt, partsNoOpt);
-      } else if (keyword.equals("hibernating")) {
+        break;
+      case "hibernating":
         this.parseHibernatingLine(line, lineNoOpt, partsNoOpt);
-      } else if (keyword.equals("uptime")) {
+        break;
+      case "uptime":
         this.parseUptimeLine(line, lineNoOpt, partsNoOpt);
-      } else if (keyword.equals("onion-key")) {
+        break;
+      case "onion-key":
         this.parseOnionKeyLine(line, lineNoOpt, partsNoOpt);
         nextCrypto = "onion-key";
-      } else if (keyword.equals("signing-key")) {
+        break;
+      case "signing-key":
         this.parseSigningKeyLine(line, lineNoOpt, partsNoOpt);
         nextCrypto = "signing-key";
-      } else if (keyword.equals("accept")) {
+        break;
+      case "accept":
         this.parseAcceptLine(line, lineNoOpt, partsNoOpt);
-      } else if (keyword.equals("reject")) {
+        break;
+      case "reject":
         this.parseRejectLine(line, lineNoOpt, partsNoOpt);
-      } else if (keyword.equals("router-signature")) {
+        break;
+      case "router-signature":
         this.parseRouterSignatureLine(line, lineNoOpt, partsNoOpt);
         nextCrypto = "router-signature";
-      } else if (keyword.equals("contact")) {
+        break;
+      case "contact":
         this.parseContactLine(line, lineNoOpt, partsNoOpt);
-      } else if (keyword.equals("family")) {
+        break;
+      case "family":
         this.parseFamilyLine(line, lineNoOpt, partsNoOpt);
-      } else if (keyword.equals("read-history")) {
+        break;
+      case "read-history":
         this.parseReadHistoryLine(line, lineNoOpt, partsNoOpt);
-      } else if (keyword.equals("write-history")) {
+        break;
+      case "write-history":
         this.parseWriteHistoryLine(line, lineNoOpt, partsNoOpt);
-      } else if (keyword.equals("eventdns")) {
+        break;
+      case "eventdns":
         this.parseEventdnsLine(line, lineNoOpt, partsNoOpt);
-      } else if (keyword.equals("caches-extra-info")) {
+        break;
+      case "caches-extra-info":
         this.parseCachesExtraInfoLine(line, lineNoOpt, partsNoOpt);
-      } else if (keyword.equals("extra-info-digest")) {
+        break;
+      case "extra-info-digest":
         this.parseExtraInfoDigestLine(line, lineNoOpt, partsNoOpt);
-      } else if (keyword.equals("hidden-service-dir")) {
+        break;
+      case "hidden-service-dir":
         this.parseHiddenServiceDirLine(line, lineNoOpt, partsNoOpt);
-      } else if (keyword.equals("protocols")) {
+        break;
+      case "protocols":
         this.parseProtocolsLine(line, lineNoOpt, partsNoOpt);
-      } else if (keyword.equals("allow-single-hop-exits")) {
+        break;
+      case "allow-single-hop-exits":
         this.parseAllowSingleHopExitsLine(line, lineNoOpt, partsNoOpt);
-      } else if (keyword.equals("dircacheport")) {
+        break;
+      case "dircacheport":
         this.parseDircacheportLine(line, lineNoOpt, partsNoOpt);
-      } else if (keyword.equals("router-digest")) {
+        break;
+      case "router-digest":
         this.parseRouterDigestLine(line, lineNoOpt, partsNoOpt);
-      } else if (keyword.equals("ipv6-policy")) {
+        break;
+      case "router-digest-sha256":
+        this.parseRouterDigestSha256Line(line, lineNoOpt, partsNoOpt);
+        break;
+      case "ipv6-policy":
         this.parseIpv6PolicyLine(line, lineNoOpt, partsNoOpt);
-      } else if (keyword.equals("ntor-onion-key")) {
+        break;
+      case "ntor-onion-key":
         this.parseNtorOnionKeyLine(line, lineNoOpt, partsNoOpt);
-      } else if (line.startsWith("-----BEGIN")) {
-        cryptoLines = new ArrayList<String>();
+        break;
+      case "identity-ed25519":
+        this.parseIdentityEd25519Line(line, lineNoOpt, partsNoOpt);
+        nextCrypto = "identity-ed25519";
+        break;
+      case "master-key-ed25519":
+        this.parseMasterKeyEd25519Line(line, lineNoOpt, partsNoOpt);
+        break;
+      case "router-sig-ed25519":
+        this.parseRouterSigEd25519Line(line, lineNoOpt, partsNoOpt);
+        break;
+      case "onion-key-crosscert":
+        this.parseOnionKeyCrosscert(line, lineNoOpt, partsNoOpt);
+        nextCrypto = "onion-key-crosscert";
+        break;
+      case "ntor-onion-key-crosscert":
+        this.parseNtorOnionKeyCrosscert(line, lineNoOpt, partsNoOpt);
+        nextCrypto = "ntor-onion-key-crosscert";
+        break;
+      case "-----BEGIN":
+        cryptoLines = new ArrayList<>();
         cryptoLines.add(line);
-      } else if (line.startsWith("-----END")) {
+        break;
+      case "-----END":
         cryptoLines.add(line);
         StringBuilder sb = new StringBuilder();
         for (String cryptoLine : cryptoLines) {
-          sb.append("\n" + cryptoLine);
+          sb.append("\n").append(cryptoLine);
         }
         String cryptoString = sb.toString().substring(1);
-        if ("onion-key".equals(nextCrypto)) {
+        switch (nextCrypto) {
+        case "onion-key":
           this.onionKey = cryptoString;
-        } else if ("signing-key".equals(nextCrypto)) {
+          break;
+        case "signing-key":
           this.signingKey = cryptoString;
-        } else if ("router-signature".equals(nextCrypto)) {
+          break;
+        case "router-signature":
           this.routerSignature = cryptoString;
-        } else if (this.failUnrecognizedDescriptorLines) {
-          throw new DescriptorParseException("Unrecognized crypto "
-              + "block '" + cryptoString + "' in server descriptor.");
-        } else {
-          if (this.unrecognizedLines == null) {
-            this.unrecognizedLines = new ArrayList<String>();
+          break;
+        case "identity-ed25519":
+          this.identityEd25519 = cryptoString;
+          this.parseIdentityEd25519CryptoBlock(cryptoString);
+          break;
+        case "onion-key-crosscert":
+          this.onionKeyCrosscert = cryptoString;
+          break;
+        case "ntor-onion-key-crosscert":
+          this.ntorOnionKeyCrosscert = cryptoString;
+          break;
+        default:
+          if (this.failUnrecognizedDescriptorLines) {
+            throw new DescriptorParseException("Unrecognized crypto "
+                + "block '" + cryptoString + "' in server descriptor.");
+          } else {
+            if (this.unrecognizedLines == null) {
+              this.unrecognizedLines = new ArrayList<>();
+            }
+            this.unrecognizedLines.addAll(cryptoLines);
           }
-          this.unrecognizedLines.addAll(cryptoLines);
         }
         cryptoLines = null;
-        nextCrypto = null;
-      } else if (cryptoLines != null) {
-        cryptoLines.add(line);
-      } else {
-        ParseHelper.parseKeyword(line, partsNoOpt[0]);
-        if (this.failUnrecognizedDescriptorLines) {
-          throw new DescriptorParseException("Unrecognized line '"
-              + line + "' in server descriptor.");
+        nextCrypto = "";
+        break;
+      default:
+        if (cryptoLines != null) {
+          cryptoLines.add(line);
         } else {
-          if (this.unrecognizedLines == null) {
-            this.unrecognizedLines = new ArrayList<String>();
+          ParseHelper.parseKeyword(line, partsNoOpt[0]);
+          if (this.failUnrecognizedDescriptorLines) {
+            throw new DescriptorParseException("Unrecognized line '"
+                + line + "' in server descriptor.");
+          } else {
+            if (this.unrecognizedLines == null) {
+              this.unrecognizedLines = new ArrayList<>();
+            }
+            this.unrecognizedLines.add(line);
           }
-          this.unrecognizedLines.add(line);
         }
       }
     }
@@ -264,13 +319,7 @@ public class ServerDescriptorImpl extends DescriptorImpl
     if (partsNoOpt.length != 2) {
       throw new DescriptorParseException("Illegal line '" + line + "'.");
     }
-    if (partsNoOpt[1].equals("1")) {
-      this.hibernating = true;
-    } else if (partsNoOpt[1].equals("0")) {
-      this.hibernating = false;
-    } else {
-      throw new DescriptorParseException("Illegal line '" + line + "'.");
-    }
+    this.hibernating = ParseHelper.parseBoolean(partsNoOpt[1], line);
   }
 
   private void parseUptimeLine(String line, String lineNoOpt,
@@ -385,13 +434,7 @@ public class ServerDescriptorImpl extends DescriptorImpl
     if (partsNoOpt.length != 2) {
       throw new DescriptorParseException("Illegal line '" + line + "'.");
     }
-    if (partsNoOpt[1].equals("1")) {
-      this.usesEnhancedDnsLogic = true;
-    } else if (partsNoOpt[1].equals("0")) {
-      this.usesEnhancedDnsLogic = false;
-    } else {
-      throw new DescriptorParseException("Illegal line '" + line + "'.");
-    }
+    this.usesEnhancedDnsLogic = ParseHelper.parseBoolean(partsNoOpt[1], line);
   }
 
   private void parseCachesExtraInfoLine(String line, String lineNoOpt,
@@ -409,6 +452,10 @@ public class ServerDescriptorImpl extends DescriptorImpl
     }
     this.extraInfoDigest = ParseHelper.parseTwentyByteHexString(line,
         partsNoOpt[1]);
+    if (partsNoOpt.length >= 3) {
+      ParseHelper.parseThirtyTwoByteBase64String(line, partsNoOpt[2]);
+      this.extraInfoDigestSha256 = partsNoOpt[2];
+    }
   }
 
   private void parseHiddenServiceDirLine(String line, String lineNoOpt,
@@ -433,10 +480,15 @@ public class ServerDescriptorImpl extends DescriptorImpl
       String[] partsNoOpt) throws DescriptorParseException {
     int linkIndex = -1, circuitIndex = -1;
     for (int i = 1; i < partsNoOpt.length; i++) {
-      if (partsNoOpt[i].equals("Link")) {
+      switch (partsNoOpt[i]) {
+      case "Link":
         linkIndex = i;
-      } else if (partsNoOpt[i].equals("Circuit")) {
+        break;
+      case "Circuit":
         circuitIndex = i;
+        break;
+      default:
+        // empty
       }
     }
     if (linkIndex < 0 || circuitIndex < 0 || circuitIndex < linkIndex) {
@@ -498,19 +550,23 @@ public class ServerDescriptorImpl extends DescriptorImpl
     boolean isValid = true;
     if (partsNoOpt.length != 3) {
       isValid = false;
-    } else if (!partsNoOpt[1].equals("accept") &&
-        !partsNoOpt[1].equals("reject")) {
-      isValid = false;
     } else {
-      this.ipv6DefaultPolicy = partsNoOpt[1];
-      this.ipv6PortList = partsNoOpt[2];
-      String[] ports = partsNoOpt[2].split(",", -1);
-      for (int i = 0; i < ports.length; i++) {
-        if (ports[i].length() < 1) {
-          isValid = false;
+        switch (partsNoOpt[1]) {
+        case "accept":
+        case "reject":
+          this.ipv6DefaultPolicy = partsNoOpt[1];
+          this.ipv6PortList = partsNoOpt[2];
+          String[] ports = partsNoOpt[2].split(",", -1);
+          for (int i = 0; i < ports.length; i++) {
+            if (ports[i].length() < 1) {
+              isValid = false;
+              break;
+            }
+          }
           break;
+        default:
+          isValid = false;
         }
-      }
     }
     if (!isValid) {
       throw new DescriptorParseException("Illegal line '" + line + "'.");
@@ -525,6 +581,76 @@ public class ServerDescriptorImpl extends DescriptorImpl
     this.ntorOnionKey = partsNoOpt[1].replaceAll("=", "");
   }
 
+  private void parseIdentityEd25519Line(String line, String lineNoOpt,
+      String[] partsNoOpt) throws DescriptorParseException {
+    if (partsNoOpt.length != 1) {
+      throw new DescriptorParseException("Illegal line '" + line + "'.");
+    }
+  }
+
+  private void parseOnionKeyCrosscert(String line, String lineNoOpt,
+      String[] partsNoOpt) throws DescriptorParseException {
+    if (partsNoOpt.length != 1) {
+      throw new DescriptorParseException("Illegal line '" + line + "'.");
+    }
+  }
+
+  private void parseNtorOnionKeyCrosscert(String line, String lineNoOpt,
+      String[] partsNoOpt) throws DescriptorParseException {
+    if (partsNoOpt.length != 2) {
+      throw new DescriptorParseException("Illegal line '" + line + "'.");
+    }
+    try {
+      this.ntorOnionKeyCrosscertSign = Integer.parseInt(partsNoOpt[1]);
+    } catch (NumberFormatException e) {
+      throw new DescriptorParseException("Illegal line '" + line + "'.");
+    }
+  }
+
+  private void parseIdentityEd25519CryptoBlock(String cryptoString)
+      throws DescriptorParseException {
+    String masterKeyEd25519FromIdentityEd25519 =
+        ParseHelper.parseMasterKeyEd25519FromIdentityEd25519CryptoBlock(
+        cryptoString);
+    if (this.masterKeyEd25519 != null && !this.masterKeyEd25519.equals(
+        masterKeyEd25519FromIdentityEd25519)) {
+      throw new DescriptorParseException("Mismatch between "
+          + "identity-ed25519 and master-key-ed25519.");
+    }
+    this.masterKeyEd25519 = masterKeyEd25519FromIdentityEd25519;
+  }
+
+  private void parseMasterKeyEd25519Line(String line, String lineNoOpt,
+      String[] partsNoOpt) throws DescriptorParseException {
+    if (partsNoOpt.length != 2) {
+      throw new DescriptorParseException("Illegal line '" + line + "'.");
+    }
+    String masterKeyEd25519FromMasterKeyEd25519Line = partsNoOpt[1];
+    if (this.masterKeyEd25519 != null && !masterKeyEd25519.equals(
+        masterKeyEd25519FromMasterKeyEd25519Line)) {
+      throw new DescriptorParseException("Mismatch between "
+          + "identity-ed25519 and master-key-ed25519.");
+    }
+    this.masterKeyEd25519 = masterKeyEd25519FromMasterKeyEd25519Line;
+  }
+
+  private void parseRouterSigEd25519Line(String line, String lineNoOpt,
+      String[] partsNoOpt) throws DescriptorParseException {
+    if (partsNoOpt.length != 2) {
+      throw new DescriptorParseException("Illegal line '" + line + "'.");
+    }
+    this.routerSignatureEd25519 = partsNoOpt[1];
+  }
+
+  private void parseRouterDigestSha256Line(String line, String lineNoOpt,
+      String[] partsNoOpt) throws DescriptorParseException {
+    if (partsNoOpt.length != 2) {
+      throw new DescriptorParseException("Illegal line '" + line + "'.");
+    }
+    ParseHelper.parseThirtyTwoByteBase64String(line, partsNoOpt[1]);
+    this.serverDescriptorDigestSha256 = partsNoOpt[1];
+  }
+
   private void calculateDigest() throws DescriptorParseException {
     if (this.serverDescriptorDigest != null) {
       /* We already learned the descriptor digest of this bridge
@@ -556,11 +682,48 @@ public class ServerDescriptorImpl extends DescriptorImpl
     }
   }
 
+  private void calculateDigestSha256() throws DescriptorParseException {
+    if (this.serverDescriptorDigestSha256 != null) {
+      /* We already learned the descriptor digest of this bridge
+       * descriptor from a "router-digest-sha256" line. */
+      return;
+    }
+    try {
+      String ascii = new String(this.getRawDescriptorBytes(), "US-ASCII");
+      String startToken = "router ";
+      String sigToken = "\n-----END SIGNATURE-----\n";
+      int start = ascii.indexOf(startToken);
+      int sig = ascii.indexOf(sigToken) + sigToken.length();
+      if (start >= 0 && sig >= 0 && sig > start) {
+        byte[] forDigest = new byte[sig - start];
+        System.arraycopy(this.getRawDescriptorBytes(), start, forDigest,
+            0, sig - start);
+        this.serverDescriptorDigestSha256 =
+            DatatypeConverter.printBase64Binary(
+            MessageDigest.getInstance("SHA-256").digest(forDigest)).
+            replaceAll("=", "");
+      }
+    } catch (UnsupportedEncodingException e) {
+      /* Handle below. */
+    } catch (NoSuchAlgorithmException e) {
+      /* Handle below. */
+    }
+    if (this.serverDescriptorDigestSha256 == null) {
+      throw new DescriptorParseException("Could not calculate server "
+          + "descriptor SHA-256 digest.");
+    }
+  }
+
   private String serverDescriptorDigest;
   public String getServerDescriptorDigest() {
     return this.serverDescriptorDigest;
   }
 
+  private String serverDescriptorDigestSha256;
+  public String getServerDescriptorDigestSha256() {
+    return this.serverDescriptorDigestSha256;
+  }
+
   private String nickname;
   public String getNickname() {
     return this.nickname;
@@ -586,9 +749,9 @@ public class ServerDescriptorImpl extends DescriptorImpl
     return this.dirPort;
   }
 
-  private List<String> orAddresses = new ArrayList<String>();
+  private List<String> orAddresses = new ArrayList<>();
   public List<String> getOrAddresses() {
-    return new ArrayList<String>(this.orAddresses);
+    return new ArrayList<>(this.orAddresses);
   }
 
   private int bandwidthRate;
@@ -641,9 +804,9 @@ public class ServerDescriptorImpl extends DescriptorImpl
     return this.signingKey;
   }
 
-  private List<String> exitPolicyLines = new ArrayList<String>();
+  private List<String> exitPolicyLines = new ArrayList<>();
   public List<String> getExitPolicyLines() {
-    return new ArrayList<String>(this.exitPolicyLines);
+    return new ArrayList<>(this.exitPolicyLines);
   }
 
   private String routerSignature;
@@ -687,6 +850,11 @@ public class ServerDescriptorImpl extends DescriptorImpl
     return this.extraInfoDigest;
   }
 
+  private String extraInfoDigestSha256;
+  public String getExtraInfoDigestSha256() {
+    return this.extraInfoDigestSha256;
+  }
+
   private Integer[] hiddenServiceDirVersions;
   public List<Integer> getHiddenServiceDirVersions() {
     return this.hiddenServiceDirVersions == null ? null :
@@ -724,5 +892,35 @@ public class ServerDescriptorImpl extends DescriptorImpl
   public String getNtorOnionKey() {
     return this.ntorOnionKey;
   }
+
+  private String identityEd25519;
+  public String getIdentityEd25519() {
+    return this.identityEd25519;
+  }
+
+  private String masterKeyEd25519;
+  public String getMasterKeyEd25519() {
+    return this.masterKeyEd25519;
+  }
+
+  private String routerSignatureEd25519;
+  public String getRouterSignatureEd25519() {
+    return this.routerSignatureEd25519;
+  }
+
+  private String onionKeyCrosscert;
+  public String getOnionKeyCrosscert() {
+    return this.onionKeyCrosscert;
+  }
+
+  private String ntorOnionKeyCrosscert;
+  public String getNtorOnionKeyCrosscert() {
+    return this.ntorOnionKeyCrosscert;
+  }
+
+  private int ntorOnionKeyCrosscertSign = -1;
+  public int getNtorOnionKeyCrosscertSign() {
+    return ntorOnionKeyCrosscertSign;
+  }
 }
 
