@@ -2,7 +2,6 @@
  * See LICENSE for licensing information */
 package org.torproject.descriptor.impl;
 
-import org.torproject.descriptor.DescriptorParseException;
 import static org.junit.Assert.assertEquals;
 import static org.junit.Assert.assertFalse;
 import static org.junit.Assert.assertNotNull;
@@ -16,7 +15,10 @@ import java.util.Map;
 import java.util.SortedMap;
 
 import org.junit.Test;
+import org.torproject.descriptor.BridgeExtraInfoDescriptor;
+import org.torproject.descriptor.DescriptorParseException;
 import org.torproject.descriptor.ExtraInfoDescriptor;
+import org.torproject.descriptor.RelayExtraInfoDescriptor;
 
 /* Test parsing of extra-info descriptors. */
 public class ExtraInfoDescriptorImplTest {
@@ -30,14 +32,14 @@ public class ExtraInfoDescriptorImplTest {
         String line) throws DescriptorParseException {
       DescriptorBuilder db = new DescriptorBuilder();
       db.extraInfoLine = line;
-      return new ExtraInfoDescriptorImpl(db.buildDescriptor(), true);
+      return new RelayExtraInfoDescriptorImpl(db.buildDescriptor(), true);
     }
     private String publishedLine = "published 2012-02-11 09:08:36";
     private static ExtraInfoDescriptor createWithPublishedLine(
         String line) throws DescriptorParseException {
       DescriptorBuilder db = new DescriptorBuilder();
       db.publishedLine = line;
-      return new ExtraInfoDescriptorImpl(db.buildDescriptor(), true);
+      return new RelayExtraInfoDescriptorImpl(db.buildDescriptor(), true);
     }
     private String writeHistoryLine = "write-history 2012-02-11 09:03:39 "
         + "(900 s) 4713350144,4723824640,4710717440,4572675072";
@@ -45,7 +47,7 @@ public class ExtraInfoDescriptorImplTest {
         String line) throws DescriptorParseException {
       DescriptorBuilder db = new DescriptorBuilder();
       db.writeHistoryLine = line;
-      return new ExtraInfoDescriptorImpl(db.buildDescriptor(), true);
+      return new RelayExtraInfoDescriptorImpl(db.buildDescriptor(), true);
     }
     private String readHistoryLine = "read-history 2012-02-11 09:03:39 "
         + "(900 s) 4707695616,4699666432,4650004480,4489718784";
@@ -53,7 +55,7 @@ public class ExtraInfoDescriptorImplTest {
         String line) throws DescriptorParseException {
       DescriptorBuilder db = new DescriptorBuilder();
       db.readHistoryLine = line;
-      return new ExtraInfoDescriptorImpl(db.buildDescriptor(), true);
+      return new RelayExtraInfoDescriptorImpl(db.buildDescriptor(), true);
     }
     private String dirreqWriteHistoryLine = "dirreq-write-history "
         + "2012-02-11 09:03:39 (900 s) 81281024,64996352,60625920,"
@@ -62,7 +64,7 @@ public class ExtraInfoDescriptorImplTest {
         String line) throws DescriptorParseException {
       DescriptorBuilder db = new DescriptorBuilder();
       db.dirreqWriteHistoryLine = line;
-      return new ExtraInfoDescriptorImpl(db.buildDescriptor(), true);
+      return new RelayExtraInfoDescriptorImpl(db.buildDescriptor(), true);
     }
     private String dirreqReadHistoryLine = "dirreq-read-history "
         + "2012-02-11 09:03:39 (900 s) 17074176,16235520,16005120,"
@@ -71,70 +73,77 @@ public class ExtraInfoDescriptorImplTest {
         String line) throws DescriptorParseException {
       DescriptorBuilder db = new DescriptorBuilder();
       db.dirreqReadHistoryLine = line;
-      return new ExtraInfoDescriptorImpl(db.buildDescriptor(), true);
+      return new RelayExtraInfoDescriptorImpl(db.buildDescriptor(), true);
     }
     private String geoipDbDigestLine = null;
     private static ExtraInfoDescriptor createWithGeoipDbDigestLine(
         String line) throws DescriptorParseException {
       DescriptorBuilder db = new DescriptorBuilder();
       db.geoipDbDigestLine = line;
-      return new ExtraInfoDescriptorImpl(db.buildDescriptor(), true);
+      return new RelayExtraInfoDescriptorImpl(db.buildDescriptor(), true);
     }
     private String geoip6DbDigestLine = null;
     private static ExtraInfoDescriptor createWithGeoip6DbDigestLine(
         String line) throws DescriptorParseException {
       DescriptorBuilder db = new DescriptorBuilder();
       db.geoip6DbDigestLine = line;
-      return new ExtraInfoDescriptorImpl(db.buildDescriptor(), true);
+      return new RelayExtraInfoDescriptorImpl(db.buildDescriptor(), true);
     }
     private String geoipStatsLines = null;
     private static ExtraInfoDescriptor createWithGeoipStatsLines(
         String lines) throws DescriptorParseException {
       DescriptorBuilder db = new DescriptorBuilder();
       db.geoipStatsLines = lines;
-      return new ExtraInfoDescriptorImpl(db.buildDescriptor(), true);
+      return new RelayExtraInfoDescriptorImpl(db.buildDescriptor(), true);
     }
     private String dirreqStatsLines = null;
     private static ExtraInfoDescriptor createWithDirreqStatsLines(
         String lines) throws DescriptorParseException {
       DescriptorBuilder db = new DescriptorBuilder();
       db.dirreqStatsLines = lines;
-      return new ExtraInfoDescriptorImpl(db.buildDescriptor(), true);
+      return new RelayExtraInfoDescriptorImpl(db.buildDescriptor(), true);
     }
     private String entryStatsLines = null;
     private static ExtraInfoDescriptor createWithEntryStatsLines(
         String lines) throws DescriptorParseException {
       DescriptorBuilder db = new DescriptorBuilder();
       db.entryStatsLines = lines;
-      return new ExtraInfoDescriptorImpl(db.buildDescriptor(), true);
+      return new RelayExtraInfoDescriptorImpl(db.buildDescriptor(), true);
     }
     private String cellStatsLines = null;
     private static ExtraInfoDescriptor createWithCellStatsLines(
         String lines) throws DescriptorParseException {
       DescriptorBuilder db = new DescriptorBuilder();
       db.cellStatsLines = lines;
-      return new ExtraInfoDescriptorImpl(db.buildDescriptor(), true);
+      return new RelayExtraInfoDescriptorImpl(db.buildDescriptor(), true);
     }
     private String connBiDirectLine = null;
     private static ExtraInfoDescriptor createWithConnBiDirectLine(
         String line) throws DescriptorParseException {
       DescriptorBuilder db = new DescriptorBuilder();
       db.connBiDirectLine = line;
-      return new ExtraInfoDescriptorImpl(db.buildDescriptor(), true);
+      return new RelayExtraInfoDescriptorImpl(db.buildDescriptor(), true);
     }
     private String exitStatsLines = null;
     private static ExtraInfoDescriptor createWithExitStatsLines(
         String lines) throws DescriptorParseException {
       DescriptorBuilder db = new DescriptorBuilder();
       db.exitStatsLines = lines;
-      return new ExtraInfoDescriptorImpl(db.buildDescriptor(), true);
+      return new RelayExtraInfoDescriptorImpl(db.buildDescriptor(), true);
     }
     private String bridgeStatsLines = null;
     private static ExtraInfoDescriptor createWithBridgeStatsLines(
         String lines) throws DescriptorParseException {
       DescriptorBuilder db = new DescriptorBuilder();
       db.bridgeStatsLines = lines;
-      return new ExtraInfoDescriptorImpl(db.buildDescriptor(), true);
+      return new RelayExtraInfoDescriptorImpl(db.buildDescriptor(), true);
+    }
+    private String hidservStatsLines = null;
+    private static ExtraInfoDescriptor createWithHidservStatsLines(
+        String lines) throws DescriptorParseException {
+      DescriptorBuilder db = new DescriptorBuilder();
+      db.hidservStatsLines = lines;
+      return new RelayExtraInfoDescriptorImpl(db.buildDescriptor(), true);
     }
     private String unrecognizedLine = null;
     private static ExtraInfoDescriptor createWithUnrecognizedLine(
@@ -142,7 +151,7 @@ public class ExtraInfoDescriptorImplTest {
         throws DescriptorParseException {
       DescriptorBuilder db = new DescriptorBuilder();
       db.unrecognizedLine = line;
-      return new ExtraInfoDescriptorImpl(db.buildDescriptor(),
+      return new RelayExtraInfoDescriptorImpl(db.buildDescriptor(),
           failUnrecognizedDescriptorLines);
     }
     private byte[] nonAsciiLineBytes = null;
@@ -151,7 +160,7 @@ public class ExtraInfoDescriptorImplTest {
         throws DescriptorParseException {
       DescriptorBuilder db = new DescriptorBuilder();
       db.nonAsciiLineBytes = lineBytes;
-      return new ExtraInfoDescriptorImpl(db.buildDescriptor(),
+      return new RelayExtraInfoDescriptorImpl(db.buildDescriptor(),
           failUnrecognizedDescriptorLines);
     }
     private String routerSignatureLines = "router-signature\n"
@@ -164,57 +173,77 @@ public class ExtraInfoDescriptorImplTest {
         String line) throws DescriptorParseException {
       DescriptorBuilder db = new DescriptorBuilder();
       db.routerSignatureLines = line;
-      return new ExtraInfoDescriptorImpl(db.buildDescriptor(), true);
+      return new RelayExtraInfoDescriptorImpl(db.buildDescriptor(), true);
+    }
+    private String identityEd25519Lines = null,
+        masterKeyEd25519Line = null, routerSigEd25519Line = null;
+    private static ExtraInfoDescriptor createWithEd25519Lines(
+        String identityEd25519Lines, String masterKeyEd25519Line,
+        String routerSigEd25519Line) throws DescriptorParseException {
+      DescriptorBuilder db = new DescriptorBuilder();
+      db.identityEd25519Lines = identityEd25519Lines;
+      db.masterKeyEd25519Line = masterKeyEd25519Line;
+      db.routerSigEd25519Line = routerSigEd25519Line;
+      return new RelayExtraInfoDescriptorImpl(db.buildDescriptor(), true);
     }
     private byte[] buildDescriptor() {
       StringBuilder sb = new StringBuilder();
       if (this.extraInfoLine != null) {
-        sb.append(this.extraInfoLine + "\n");
+        sb.append(this.extraInfoLine).append("\n");
+      }
+      if (this.identityEd25519Lines != null) {
+        sb.append(this.identityEd25519Lines).append("\n");
+      }
+      if (this.masterKeyEd25519Line != null) {
+        sb.append(this.masterKeyEd25519Line).append("\n");
       }
       if (this.publishedLine != null) {
-        sb.append(this.publishedLine + "\n");
+        sb.append(this.publishedLine).append("\n");
       }
       if (this.writeHistoryLine != null) {
-        sb.append(this.writeHistoryLine + "\n");
+        sb.append(this.writeHistoryLine).append("\n");
       }
       if (this.readHistoryLine != null) {
-        sb.append(this.readHistoryLine + "\n");
+        sb.append(this.readHistoryLine).append("\n");
       }
       if (this.dirreqWriteHistoryLine != null) {
-        sb.append(this.dirreqWriteHistoryLine + "\n");
+        sb.append(this.dirreqWriteHistoryLine).append("\n");
       }
       if (this.dirreqReadHistoryLine != null) {
-        sb.append(this.dirreqReadHistoryLine + "\n");
+        sb.append(this.dirreqReadHistoryLine).append("\n");
       }
       if (this.geoipDbDigestLine != null) {
-        sb.append(this.geoipDbDigestLine + "\n");
+        sb.append(this.geoipDbDigestLine).append("\n");
       }
       if (this.geoip6DbDigestLine != null) {
-        sb.append(this.geoip6DbDigestLine + "\n");
+        sb.append(this.geoip6DbDigestLine).append("\n");
       }
       if (this.geoipStatsLines != null) {
-        sb.append(this.geoipStatsLines + "\n");
+        sb.append(this.geoipStatsLines).append("\n");
       }
       if (this.dirreqStatsLines != null) {
-        sb.append(this.dirreqStatsLines + "\n");
+        sb.append(this.dirreqStatsLines).append("\n");
       }
       if (this.entryStatsLines != null) {
-        sb.append(this.entryStatsLines + "\n");
+        sb.append(this.entryStatsLines).append("\n");
       }
       if (this.cellStatsLines != null) {
-        sb.append(this.cellStatsLines + "\n");
+        sb.append(this.cellStatsLines).append("\n");
       }
       if (this.connBiDirectLine != null) {
-        sb.append(this.connBiDirectLine + "\n");
+        sb.append(this.connBiDirectLine).append("\n");
       }
       if (this.exitStatsLines != null) {
-        sb.append(this.exitStatsLines + "\n");
+        sb.append(this.exitStatsLines).append("\n");
       }
       if (this.bridgeStatsLines != null) {
-        sb.append(this.bridgeStatsLines + "\n");
+        sb.append(this.bridgeStatsLines).append("\n");
+      }
+      if (this.hidservStatsLines != null) {
+        sb.append(this.hidservStatsLines).append("\n");
       }
       if (this.unrecognizedLine != null) {
-        sb.append(this.unrecognizedLine + "\n");
+        sb.append(this.unrecognizedLine).append("\n");
       }
       if (this.nonAsciiLineBytes != null) {
         try {
@@ -230,8 +259,11 @@ public class ExtraInfoDescriptorImplTest {
           return null;
         }
       }
+      if (this.routerSigEd25519Line != null) {
+        sb.append(this.routerSigEd25519Line).append("\n");
+      }
       if (this.routerSignatureLines != null) {
-        sb.append(this.routerSignatureLines + "\n");
+        sb.append(this.routerSignatureLines).append("\n");
       }
       return sb.toString().getBytes();
     }
@@ -266,10 +298,10 @@ public class ExtraInfoDescriptorImplTest {
     private String buildGeoipStatsLines() {
       StringBuilder sb = new StringBuilder();
       if (this.geoipStartTimeLine != null) {
-        sb.append(this.geoipStartTimeLine + "\n");
+        sb.append(this.geoipStartTimeLine).append("\n");
       }
       if (this.geoipClientOriginsLine != null) {
-        sb.append(this.geoipClientOriginsLine + "\n");
+        sb.append(this.geoipClientOriginsLine).append("\n");
       }
       String lines = sb.toString();
       if (lines.endsWith("\n")) {
@@ -408,43 +440,43 @@ public class ExtraInfoDescriptorImplTest {
     private String buildDirreqStatsLines() {
       StringBuilder sb = new StringBuilder();
       if (this.dirreqStatsEndLine != null) {
-        sb.append(this.dirreqStatsEndLine + "\n");
+        sb.append(this.dirreqStatsEndLine).append("\n");
       }
       if (this.dirreqV3IpsLine != null) {
-        sb.append(this.dirreqV3IpsLine + "\n");
+        sb.append(this.dirreqV3IpsLine).append("\n");
       }
       if (this.dirreqV2IpsLine != null) {
-        sb.append(this.dirreqV2IpsLine + "\n");
+        sb.append(this.dirreqV2IpsLine).append("\n");
       }
       if (this.dirreqV3ReqsLine != null) {
-        sb.append(this.dirreqV3ReqsLine + "\n");
+        sb.append(this.dirreqV3ReqsLine).append("\n");
       }
       if (this.dirreqV2ReqsLine != null) {
-        sb.append(this.dirreqV2ReqsLine + "\n");
+        sb.append(this.dirreqV2ReqsLine).append("\n");
       }
       if (this.dirreqV3RespLine != null) {
-        sb.append(this.dirreqV3RespLine + "\n");
+        sb.append(this.dirreqV3RespLine).append("\n");
       }
       if (this.dirreqV2RespLine != null) {
-        sb.append(this.dirreqV2RespLine + "\n");
+        sb.append(this.dirreqV2RespLine).append("\n");
       }
       if (this.dirreqV2ShareLine != null) {
-        sb.append(this.dirreqV2ShareLine + "\n");
+        sb.append(this.dirreqV2ShareLine).append("\n");
       }
       if (this.dirreqV3ShareLine != null) {
-        sb.append(this.dirreqV3ShareLine + "\n");
+        sb.append(this.dirreqV3ShareLine).append("\n");
       }
       if (this.dirreqV3DirectDlLine != null) {
-        sb.append(this.dirreqV3DirectDlLine + "\n");
+        sb.append(this.dirreqV3DirectDlLine).append("\n");
       }
       if (this.dirreqV2DirectDlLine != null) {
-        sb.append(this.dirreqV2DirectDlLine + "\n");
+        sb.append(this.dirreqV2DirectDlLine).append("\n");
       }
       if (this.dirreqV3TunneledDlLine != null) {
-        sb.append(this.dirreqV3TunneledDlLine + "\n");
+        sb.append(this.dirreqV3TunneledDlLine).append("\n");
       }
       if (this.dirreqV2TunneledDlLine != null) {
-        sb.append(this.dirreqV2TunneledDlLine + "\n");
+        sb.append(this.dirreqV2TunneledDlLine).append("\n");
       }
       String lines = sb.toString();
       if (lines.endsWith("\n")) {
@@ -484,10 +516,10 @@ public class ExtraInfoDescriptorImplTest {
     private String buildEntryStatsLines() {
       StringBuilder sb = new StringBuilder();
       if (this.entryStatsEndLine != null) {
-        sb.append(this.entryStatsEndLine + "\n");
+        sb.append(this.entryStatsEndLine).append("\n");
       }
       if (this.entryIpsLine != null) {
-        sb.append(this.entryIpsLine + "\n");
+        sb.append(this.entryIpsLine).append("\n");
       }
       String lines = sb.toString();
       if (lines.endsWith("\n")) {
@@ -554,19 +586,19 @@ public class ExtraInfoDescriptorImplTest {
     private String buildCellStatsLines() {
       StringBuilder sb = new StringBuilder();
       if (this.cellStatsEndLine != null) {
-        sb.append(this.cellStatsEndLine + "\n");
+        sb.append(this.cellStatsEndLine).append("\n");
       }
       if (this.cellProcessedCellsLine != null) {
-        sb.append(this.cellProcessedCellsLine + "\n");
+        sb.append(this.cellProcessedCellsLine).append("\n");
       }
       if (this.cellQueuedCellsLine != null) {
-        sb.append(this.cellQueuedCellsLine + "\n");
+        sb.append(this.cellQueuedCellsLine).append("\n");
       }
       if (this.cellTimeInQueueLine != null) {
-        sb.append(this.cellTimeInQueueLine + "\n");
+        sb.append(this.cellTimeInQueueLine).append("\n");
       }
       if (this.cellCircuitsPerDecileLine != null) {
-        sb.append(this.cellCircuitsPerDecileLine + "\n");
+        sb.append(this.cellCircuitsPerDecileLine).append("\n");
       }
       String lines = sb.toString();
       if (lines.endsWith("\n")) {
@@ -627,16 +659,16 @@ public class ExtraInfoDescriptorImplTest {
     private String buildExitStatsLines() {
       StringBuilder sb = new StringBuilder();
       if (this.exitStatsEndLine != null) {
-        sb.append(this.exitStatsEndLine + "\n");
+        sb.append(this.exitStatsEndLine).append("\n");
       }
       if (this.exitKibibytesWrittenLine != null) {
-        sb.append(this.exitKibibytesWrittenLine + "\n");
+        sb.append(this.exitKibibytesWrittenLine).append("\n");
       }
       if (this.exitKibibytesReadLine != null) {
-        sb.append(this.exitKibibytesReadLine + "\n");
+        sb.append(this.exitKibibytesReadLine).append("\n");
       }
       if (this.exitStreamsOpenedLine != null) {
-        sb.append(this.exitStreamsOpenedLine + "\n");
+        sb.append(this.exitStreamsOpenedLine).append("\n");
       }
       String lines = sb.toString();
       if (lines.endsWith("\n")) {
@@ -692,16 +724,72 @@ public class ExtraInfoDescriptorImplTest {
     private String buildBridgeStatsLines() {
       StringBuilder sb = new StringBuilder();
       if (this.bridgeStatsEndLine != null) {
-        sb.append(this.bridgeStatsEndLine + "\n");
+        sb.append(this.bridgeStatsEndLine).append("\n");
       }
       if (this.bridgeIpsLine != null) {
-        sb.append(this.bridgeIpsLine + "\n");
+        sb.append(this.bridgeIpsLine).append("\n");
       }
       if (this.bridgeIpVersionsLine != null) {
-        sb.append(this.bridgeIpVersionsLine + "\n");
+        sb.append(this.bridgeIpVersionsLine).append("\n");
       }
       if (this.bridgeIpTransportsLine != null) {
-        sb.append(this.bridgeIpTransportsLine + "\n");
+        sb.append(this.bridgeIpTransportsLine).append("\n");
+      }
+      String lines = sb.toString();
+      if (lines.endsWith("\n")) {
+        lines = lines.substring(0, lines.length() - 1);
+      }
+      return lines;
+    }
+  }
+
+  /* Helper class to build a set of hidserv-stats lines based on default
+   * data and modifications requested by test methods. */
+  private static class HidservStatsBuilder {
+    private String hidservStatsEndLine = "hidserv-stats-end 2015-12-03 "
+        + "14:26:56 (86400 s)";
+    private static ExtraInfoDescriptor createWithHidservStatsEndLine(
+        String line) throws DescriptorParseException {
+      HidservStatsBuilder hsb = new HidservStatsBuilder();
+      hsb.hidservStatsEndLine = line;
+      return DescriptorBuilder.createWithHidservStatsLines(
+          hsb.buildHidservStatsLines());
+    }
+    private String hidservRendRelayedCellsLine =
+        "hidserv-rend-relayed-cells 36474281 delta_f=2048 epsilon=0.30 "
+        + "bin_size=1024";
+    private static ExtraInfoDescriptor
+        createWithHidservRendRelayedCellsLine(String line)
+        throws DescriptorParseException {
+      HidservStatsBuilder hsb = new HidservStatsBuilder();
+      hsb.hidservRendRelayedCellsLine = line;
+      return DescriptorBuilder.createWithHidservStatsLines(
+          hsb.buildHidservStatsLines());
+    }
+    private String hidservDirOnionsSeenLine = "hidserv-dir-onions-seen "
+        + "-3 delta_f=8 epsilon=0.30 bin_size=8";
+    private static ExtraInfoDescriptor createWithHidservDirOnionsSeenLine(
+        String line) throws DescriptorParseException {
+      HidservStatsBuilder hsb = new HidservStatsBuilder();
+      hsb.hidservDirOnionsSeenLine = line;
+      return DescriptorBuilder.createWithHidservStatsLines(
+          hsb.buildHidservStatsLines());
+    }
+    private static ExtraInfoDescriptor createWithDefaultLines()
+        throws DescriptorParseException {
+      return DescriptorBuilder.createWithHidservStatsLines(
+          new HidservStatsBuilder().buildHidservStatsLines());
+    }
+    private String buildHidservStatsLines() {
+      StringBuilder sb = new StringBuilder();
+      if (this.hidservStatsEndLine != null) {
+        sb.append(this.hidservStatsEndLine).append("\n");
+      }
+      if (this.hidservRendRelayedCellsLine != null) {
+        sb.append(this.hidservRendRelayedCellsLine).append("\n");
+      }
+      if (this.hidservDirOnionsSeenLine != null) {
+        sb.append(this.hidservDirOnionsSeenLine).append("\n");
       }
       String lines = sb.toString();
       if (lines.endsWith("\n")) {
@@ -715,7 +803,7 @@ public class ExtraInfoDescriptorImplTest {
   public void testSampleDescriptor() throws DescriptorParseException {
     DescriptorBuilder db = new DescriptorBuilder();
     ExtraInfoDescriptor descriptor =
-        new ExtraInfoDescriptorImpl(db.buildDescriptor(), true);
+        new RelayExtraInfoDescriptorImpl(db.buildDescriptor(), true);
     assertEquals("chaoscomputerclub5", descriptor.getNickname());
     assertEquals("A9C039A5FD02FCA06303DCFAABE25C5912C63B26",
         descriptor.getFingerprint());
@@ -1392,6 +1480,69 @@ public class ExtraInfoDescriptorImplTest {
         "bridge-ip-transports meek=32,obfs3_websocket=8,websocket=64");
   }
 
+  @Test()
+  public void testHidservStatsValid() throws DescriptorParseException {
+    ExtraInfoDescriptor descriptor = HidservStatsBuilder.
+        createWithDefaultLines();
+    assertEquals(1449152816000L, descriptor.getHidservStatsEndMillis());
+    assertEquals(86400L, descriptor.getHidservStatsIntervalLength());
+    assertEquals(36474281.0, descriptor.getHidservRendRelayedCells(),
+        0.0001);
+    Map<String, Double> params =
+        descriptor.getHidservRendRelayedCellsParameters();
+    assertTrue(params.containsKey("delta_f"));
+    assertEquals(2048.0, params.remove("delta_f"), 0.0001);
+    assertTrue(params.containsKey("epsilon"));
+    assertEquals(0.3, params.remove("epsilon"), 0.0001);
+    assertTrue(params.containsKey("bin_size"));
+    assertEquals(1024.0, params.remove("bin_size"), 0.0001);
+    assertTrue(params.isEmpty());
+    assertEquals(-3.0, descriptor.getHidservDirOnionsSeen(), 0.0001);
+    params = descriptor.getHidservDirOnionsSeenParameters();
+    assertTrue(params.containsKey("delta_f"));
+    assertEquals(8.0, params.remove("delta_f"), 0.0001);
+    assertTrue(params.containsKey("epsilon"));
+    assertEquals(0.3, params.remove("epsilon"), 0.0001);
+    assertTrue(params.containsKey("bin_size"));
+    assertEquals(8.0, params.remove("bin_size"), 0.0001);
+    assertTrue(params.isEmpty());
+  }
+
+  @Test()
+  public void testHidservStatsEndLineMissing()
+      throws DescriptorParseException {
+    ExtraInfoDescriptor descriptor =
+        HidservStatsBuilder.createWithHidservStatsEndLine(null);
+    assertEquals(-1L, descriptor.getHidservStatsEndMillis());
+    assertEquals(-1L, descriptor.getHidservStatsIntervalLength());
+  }
+
+  @Test()
+  public void testHidservRendRelayedCellsNoParams()
+        throws DescriptorParseException {
+    ExtraInfoDescriptor descriptor =
+        HidservStatsBuilder.createWithHidservRendRelayedCellsLine(
+        "hidserv-rend-relayed-cells 36474281");
+    assertEquals(36474281.0, descriptor.getHidservRendRelayedCells(),
+        0.0001);
+    assertTrue(
+        descriptor.getHidservRendRelayedCellsParameters().isEmpty());
+  }
+
+  @Test(expected = DescriptorParseException.class)
+  public void testHidservDirOnionsSeenCommaSeparatedParams()
+      throws DescriptorParseException {
+    HidservStatsBuilder.createWithHidservDirOnionsSeenLine(
+        "hidserv-dir-onions-seen -3 delta_f=8,epsilon=0.30,bin_size=8");
+  }
+
+  @Test(expected = DescriptorParseException.class)
+  public void testHidservDirOnionsSeenNoDoubleParams()
+      throws DescriptorParseException {
+    HidservStatsBuilder.createWithHidservDirOnionsSeenLine(
+        "hidserv-dir-onions-seen -3 delta_f=A epsilon=B bin_size=C");
+  }
+
   @Test()
   public void testRouterSignatureOpt()
       throws DescriptorParseException {
@@ -1426,9 +1577,161 @@ public class ExtraInfoDescriptorImplTest {
     String unrecognizedLine = "unrecognized-line 1";
     ExtraInfoDescriptor descriptor = DescriptorBuilder.
         createWithUnrecognizedLine(unrecognizedLine, false);
-    List<String> unrecognizedLines = new ArrayList<String>();
+    List<String> unrecognizedLines = new ArrayList<>();
     unrecognizedLines.add(unrecognizedLine);
     assertEquals(unrecognizedLines, descriptor.getUnrecognizedLines());
   }
+
+  private static final String IDENTITY_ED25519_LINES =
+      "identity-ed25519\n"
+      + "-----BEGIN ED25519 CERT-----\n"
+      + "AQQABiX1AVGv5BuzJroQXbOh6vv1nbwc5rh2S13PyRFuLhTiifK4AQAgBACBCMwr"
+      + "\n4qgIlFDIzoC9ieJOtSkwrK+yXJPKlP8ojvgkx8cGKvhokOwA1eYDombzfwHcJ1"
+      + "EV\nbhEn/6g8i7wzO3LoqefIUrSAeEExOAOmm5mNmUIzL8EtnT6JHCr/sqUTUgA="
+      + "\n"
+      + "-----END ED25519 CERT-----";
+
+  private static final String MASTER_KEY_ED25519_LINE =
+      "master-key-ed25519 gQjMK+KoCJRQyM6AvYniTrUpMKyvslyTypT/KI74JMc";
+
+  private static final String ROUTER_SIG_ED25519_LINE =
+      "router-sig-ed25519 y7WF9T2GFwkSDPZEhB55HgquIFOl5uXUFMYJPq3CXXUTKeJ"
+      + "kSrtaZUB5s34fWdHQNtl84mH4dVaFMunHnwgYAw";
+
+  @Test()
+  public void testEd25519() throws DescriptorParseException {
+    ExtraInfoDescriptor descriptor =
+        DescriptorBuilder.createWithEd25519Lines(IDENTITY_ED25519_LINES,
+        MASTER_KEY_ED25519_LINE, ROUTER_SIG_ED25519_LINE);
+    assertEquals(IDENTITY_ED25519_LINES.substring(
+        IDENTITY_ED25519_LINES.indexOf("\n") + 1),
+        descriptor.getIdentityEd25519());
+    assertEquals(MASTER_KEY_ED25519_LINE.substring(
+        MASTER_KEY_ED25519_LINE.indexOf(" ") + 1),
+        descriptor.getMasterKeyEd25519());
+    assertEquals(ROUTER_SIG_ED25519_LINE.substring(
+        ROUTER_SIG_ED25519_LINE.indexOf(" ") + 1),
+        descriptor.getRouterSignatureEd25519());
+  }
+
+  @Test(expected = DescriptorParseException.class)
+  public void testEd25519IdentityMasterKeyMismatch()
+      throws DescriptorParseException {
+    DescriptorBuilder.createWithEd25519Lines(IDENTITY_ED25519_LINES,
+        "master-key-ed25519 AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
+        ROUTER_SIG_ED25519_LINE);
+  }
+
+  @Test()
+  public void testEd25519IdentityMissing()
+      throws DescriptorParseException {
+    DescriptorBuilder.createWithEd25519Lines(null,
+        MASTER_KEY_ED25519_LINE, ROUTER_SIG_ED25519_LINE);
+  }
+
+  @Test(expected = DescriptorParseException.class)
+  public void testEd25519IdentityDuplicate()
+      throws DescriptorParseException {
+    DescriptorBuilder.createWithEd25519Lines(IDENTITY_ED25519_LINES + "\n"
+        + IDENTITY_ED25519_LINES, MASTER_KEY_ED25519_LINE,
+        ROUTER_SIG_ED25519_LINE);
+  }
+
+  @Test(expected = DescriptorParseException.class)
+  public void testEd25519IdentityEmptyCrypto()
+      throws DescriptorParseException {
+    DescriptorBuilder.createWithEd25519Lines("identity-ed25519\n"
+        + "-----BEGIN ED25519 CERT-----\n-----END ED25519 CERT-----",
+        MASTER_KEY_ED25519_LINE, ROUTER_SIG_ED25519_LINE);
+  }
+
+  @Test()
+  public void testEd25519MasterKeyMissing()
+      throws DescriptorParseException {
+    ExtraInfoDescriptor descriptor =
+        DescriptorBuilder.createWithEd25519Lines(IDENTITY_ED25519_LINES,
+        null, ROUTER_SIG_ED25519_LINE);
+    assertEquals(MASTER_KEY_ED25519_LINE.substring(
+        MASTER_KEY_ED25519_LINE.indexOf(" ") + 1),
+        descriptor.getMasterKeyEd25519());
+  }
+
+  @Test(expected = DescriptorParseException.class)
+  public void testEd25519MasterKeyDuplicate()
+      throws DescriptorParseException {
+    DescriptorBuilder.createWithEd25519Lines(IDENTITY_ED25519_LINES,
+        MASTER_KEY_ED25519_LINE + "\n" + MASTER_KEY_ED25519_LINE,
+        ROUTER_SIG_ED25519_LINE);
+  }
+
+  @Test()
+  public void testEd25519RouterSigMissing()
+      throws DescriptorParseException {
+    DescriptorBuilder.createWithEd25519Lines(IDENTITY_ED25519_LINES,
+        MASTER_KEY_ED25519_LINE, null);
+  }
+
+  @Test(expected = DescriptorParseException.class)
+  public void testEd25519RouterSigDuplicate()
+      throws DescriptorParseException {
+    DescriptorBuilder.createWithEd25519Lines(IDENTITY_ED25519_LINES,
+        MASTER_KEY_ED25519_LINE, ROUTER_SIG_ED25519_LINE + "\n"
+        + ROUTER_SIG_ED25519_LINE);
+  }
+
+  @Test()
+  public void testExtraInfoDigestSha256Relay()
+      throws DescriptorParseException {
+    byte[] descriptorBytes = ("extra-info Unnamed "
+        + "EA5B335055D2F03013FF030381F02B1C631EC723\n"
+        + "identity-ed25519\n"
+        + "-----BEGIN ED25519 CERT-----\n"
+        + "AQQABiZRAenzZorGtx6xapoEeaqcLLOk3uWwJXTvOVLluSXXbRSZAQAgBADLN5"
+        + "wp\nCEOrRbshSbj1NDAUgc6cxU65M/Vx1x+b5+EXbkQZ5uiyB4pphVF5kPPT1P"
+        + "SleYqM\n8j+tlKh2i6+Xr0xScSPpmtG00/D0MoRlT7ZdaaaT5iw1DWDQCZ8BHG"
+        + "lAZwU=\n"
+        + "-----END ED25519 CERT-----\n"
+        + "published 2015-12-01 04:38:12\n"
+        + "write-history 2015-12-01 01:40:37 (14400 s) 88704000,60825600,"
+        + "61747200,76953600,61516800,59443200\n"
+        + "read-history 2015-12-01 01:40:37 (14400 s) 87321600,59443200,"
+        + "59904000,74880000,60364800,58060800\n"
+        + "router-sig-ed25519 c6eUeJs/SVjun3JhmjByEeWdRDyunSMAnGVhx71JiRj"
+        + "YzR8x5IcPebylG7m10wiolFxinvw78UhrrGo9Sq5ZBw\n"
+        + "router-signature\n"
+        + "-----BEGIN SIGNATURE-----\n"
+        + "oC2qFHCDOKSRoIPR86jdRxEYia390Z4d8fT0yr/1mg4RQ7lHmxlzFT6QxCswdX"
+        + "Ry\nvGNGR0wARySgyE+YKKWYn/Hp547JhhWd9Oc7BuFMY0XMvl/HOo+B9VjW+l"
+        + "nv6UBE\niqxx3C3Iw0ymohvOenyCUa/7TmsT7eVotDO57uIoGEc=\n"
+        + "-----END SIGNATURE-----\n"
+        + "").getBytes();
+    RelayExtraInfoDescriptor descriptor =
+        new RelayExtraInfoDescriptorImpl(descriptorBytes, true);
+    assertEquals("Pt1BtzfRwhYqGCDo8jjchS8nJP3ovrDyHGn+dqPbMgw",
+        descriptor.getExtraInfoDigestSha256());
+  }
+
+  @Test()
+  public void testExtraInfoDigestSha256Bridge()
+      throws DescriptorParseException {
+    byte[] descriptorBytes = ("extra-info idideditheconfig "
+        + "DC28749EC9E26E61DE492E46CD830379E9931B09\n"
+        + "master-key-ed25519 "
+        + "38FzmOIE6Mm85Ytx0MhFM6X9EuxWRUgb6HjyMGuO2AU\n"
+        + "published 2015-12-03 13:23:19\n"
+        + "write-history 2015-12-03 09:59:32 (14400 s) 53913600,52992000,"
+        + "53222400,53222400,53452800,53222400\n"
+        + "read-history 2015-12-03 09:59:32 (14400 s) 61056000,60364800,"
+        + "60364800,60134400,60595200,60364800\n"
+        + "geoip-db-digest 5BF366AD4A0572D82A1A0F6628AF8EF7725E3AB9\n"
+        + "geoip6-db-digest 212DE17D5A368DCAFA19B95F168BFFA101145A93\n"
+        + "router-digest-sha256 "
+        + "TvrqpjI7OmCtwGwair/NHUxg5ROVVQYz6/EDyXsDHR4\n"
+        + "router-digest 00B98F076B586272C3172B7F3DA29ADEE75F2ED8\n").getBytes();
+    BridgeExtraInfoDescriptor descriptor =
+        new BridgeExtraInfoDescriptorImpl(descriptorBytes, true);
+    assertEquals("TvrqpjI7OmCtwGwair/NHUxg5ROVVQYz6/EDyXsDHR4",
+        descriptor.getExtraInfoDigestSha256());
+  }
 }
 
