@@ -299,6 +299,8 @@ public class RelayNetworkStatusConsensusImplTest {
     assertTrue(consensus.getKnownFlags().contains("Running"));
     assertEquals(30000, (int) consensus.getConsensusParams().get(
         "CircuitPriorityHalflifeMsec"));
+    assertEquals("86.59.21.38", consensus.getDirSourceEntries().get(
+        "14C131DFC5C6F93646BE72FA1401C02A8DF2E8B4").getHostname());
     assertEquals("86.59.21.38", consensus.getDirSourceEntries().get(
         "14C131DFC5C6F93646BE72FA1401C02A8DF2E8B4").getIp());
     assertTrue(consensus.containsStatusEntry(
@@ -722,6 +724,12 @@ public class RelayNetworkStatusConsensusImplTest {
         + "4BB25CEF515B226ED03BB616EB2F60BEC8");
   }
 
+  @Test(expected = DescriptorParseException.class)
+  public void testDirSourceHostnameMissing()
+      throws DescriptorParseException {
+    DirSourceBuilder.createWithHostName("");
+  }
+
   @Test(expected = DescriptorParseException.class)
   public void testDirSourceAddress24() throws DescriptorParseException {
     DirSourceBuilder.createWithAddress("212.112.245");
